@@ -28,6 +28,8 @@ import numpy as np
 from dython.nominal import compute_associations
 # model used for feature importance, Shapley values are builtin
 import lightgbm as lgb
+from lightgbm import early_stopping
+from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit
 # visualizations
 import matplotlib as mpl
 import matplotlib.pyplot as plt
@@ -52,24 +54,22 @@ hv.renderer('bokeh').theme = 'light_minimal'
 #####################
 
 def reset_plot():
-    """
-    reset matplotlib default
-    :return: None
+    """Reset plot style
     """
     plt.rcParams = plt.rcParamsDefault
 
 
 def set_my_plt_style(height=3, width=5, linewidth=2):
-    """
-    This set the style of matplotlib to fivethirtyeight with some modifications (colours, axes)
+    """This set the style of matplotlib to fivethirtyeight with some modifications (colours, axes)
 
-    :param linewidth: int, default=2
+    Parameters
+    ----------
+    linewidth: int, default=2
         line width
-    :param height: int, default=3
+    height: int, default=3
         fig height in inches (yeah they're still struggling with the metric system)
-    :param width: int, default=5
+    width: int, default=5
         fig width in inches (yeah they're still struggling with the metric system)
-    :return: Nothing
     """
     plt.style.use('fivethirtyeight')
     my_colors_list = Bold_10.hex_colors
@@ -85,31 +85,34 @@ def set_my_plt_style(height=3, width=5, linewidth=2):
     mpl.rcParams.update(params)
 
 
-def cat_var(df, col_excl=None, return_cat=True):
-    """
-    Categorical encoding (as integer). Automatically detect the non-numerical columns,
+def cat_var(data, col_excl=None, return_cat=True):
+    """Categorical encoding (as integer). Automatically detect the non-numerical columns,
     save the index and name of those columns, encode them as integer,
     save the direct and inverse mappers as
     dictionaries.
     Return the data-set with the encoded columns with a data type either int or pandas categorical.
 
-    :param df: pd.DataFrame
+    Parameters
+    ----------
+    data: pd.DataFrame
         the dataset
-    :param col_excl: list of str, default=None
+    col_excl: list of str, default=None
         the list of columns names not being encoded (e.g. the ID column)
-    :param return_cat: bool, default=True
+    return_cat: bool, default=True
         return encoded object columns as pandas categoricals or not.
-    :return:
-     df: pd.DataFrame
+    Returns
+    -------
+    df: pd.DataFrame
         the dataframe with encoded columns
-     cat_var_df: pd.DataFrame
+    cat_var_df: pd.DataFrame
         the dataframe with the indices and names of the categorical columns
-     inv_mapper: dict
+    inv_mapper: dict
         the dictionary to map integer --> category
-     mapper: dict
+    mapper: dict
         the dictionary to map category --> integer
     """
 
+    df = data.copy()
     if col_excl is None:
         non_num_cols = list(set(list(df.columns)) - set(list(df.select_dtypes(include=[np.number]))))
     else:
@@ -123,7 +126,7 @@ def cat_var(df, col_excl=None, return_cat=True):
 
     # avoid having datetime objects as keys in the mapping dic
     date_cols = [s for s in list(df) if "date" in s]
-    df[date_cols] = df[date_cols].astype(str)
+    df.loc[:, date_cols] = df.loc[:, date_cols].astype(str)
 
     cols_need_mapped = cat_var_df.cat_name.to_list()
     inv_mapper = {col: dict(enumerate(df[col].astype('category').cat.categories))
@@ -133,7 +136,7 @@ def cat_var(df, col_excl=None, return_cat=True):
     progress_bar = tqdm(cols_need_mapped)
     for c in progress_bar:
         progress_bar.set_description('Processing {0:<30}'.format(c))
-        df[c] = df[c].map(mapper[c]).fillna(0).astype(int)
+        df.loc[:, c] = df.loc[:, c].map(mapper[c]).fillna(0).astype(int)
         # I could have use df[c].update(df[c].map(mapper[c])) while slower,
         # prevents values not included in an incomplete map from being changed to nans.
         # But then I could have outputs
@@ -142,17 +145,25 @@ def cat_var(df, col_excl=None, return_cat=True):
         # Map is faster than replace
 
     if return_cat:
-        df[non_num_cols] = df[non_num_cols].astype('category')
+        df.loc[:, non_num_cols] = df.loc[:, non_num_cols].astype('category')
     return df, cat_var_df, inv_mapper, mapper
 
+    
 
 def plot_corr(df, size=10):
-    """
-    Plot a graphical correlation matrix for a dataframe.
+    """Plot a graphical correlation matrix for a dataframe.
 
-    Input:
-        df: pandas DataFrame
-        size: vertical and horizontal size of the plot
+    Parameters
+    ----------
+    df : pd.DataFrame
+        the data
+    size : int, optional
+        figure size in inches, by default 10
+
+    Returns
+    -------
+    fig : plt.figure
+        the correlation matrix figure
     """
     set_my_plt_style()
     corr = df.corr()
@@ -177,6 +188,27 @@ def plot_corr(df, size=10):
 
 
 def plot_associations(df, features=None, size=1200, theil_u=False):
+    """Plot associations, equivalent of correlation but for
+    continuous-continuous, categorical-continuous, categorical-categorical
+    variables. 
+    
+
+    Parameters
+    ----------
+    df : pd.DataFrame
+        the data
+    features : list of str, optional
+        the list of the features to consider, if none is given all the columns are used, by default None
+    size : int, optional
+        The rendered size, in pixels, by default 1200
+    theil_u : bool, optional
+        consider or not the Theyl's U statistics for cat-cat association, by default False
+
+    Returns
+    -------
+    panel_layout : panel
+        the interactive plot
+    """
 
     if features is None:
         features = df.columns
@@ -227,8 +259,7 @@ def plot_associations(df, features=None, size=1200, theil_u=False):
 
 
 class FeatureSelector:
-    """
-    Class for performing feature selection for machine learning or data preprocessing.
+    """Class for performing feature selection for machine learning or data preprocessing.
     Heavily inspired from https://github.com/WillKoehrsen
 
     Implements five different methods to identify features for removal
@@ -336,8 +367,8 @@ class FeatureSelector:
         the dataframe with the tag "to drop" or "to keep" (== not to drop)
 
 
-    Methods:
-    --------
+    Methods
+    -------
     identify_patterns(patterns=None):
         Drop the columns by identifying patterns in their name
 
@@ -399,7 +430,7 @@ class FeatureSelector:
         `threshold` cumulative importance.
 
 
-    Example:
+    Examples
     --------
 
     # Initiate an instance of the feature selector
@@ -507,6 +538,7 @@ class FeatureSelector:
         self.mapper = None
         self.collinear_method = None
         self.tag_df = pd.DataFrame({'predictor': self.base_features})
+        self.stratified = False
 
         # Dictionary to hold removal operations
         self.ops = {}
@@ -518,13 +550,12 @@ class FeatureSelector:
         return s
 
     def identify_patterns(self, patterns=None):
-        """
-        Drop the columns by identifying patterns in their name
-        :param patterns: str
+        """Drop the columns by identifying patterns in their name
+        
+        Parameters
+        ----------
+        patterns: str
             the pattern to look for e.g. emb_ or bel_
-        :return:
-         self : object
-            Nothing but attributes
         """
         if patterns is None:
             patterns = ['_Prop']
@@ -537,12 +568,11 @@ class FeatureSelector:
         print('%d features with the pattern(s). \n' % len(self.ops['pattern']))
 
     def identify_missing(self, missing_threshold=0.1):
-        """
-        Find the features with a fraction of missing values above `missing_threshold`
-        :param missing_threshold: float, between 0 and 1. Default=0.1
-        :return:
-         self : object
-            Nothing but attributes
+        """Find the features with a fraction of missing values above `missing_threshold`
+        
+        Parameters
+        ----------
+        missing_threshold: float, between 0 and 1. Default=0.1
         """
 
         self.missing_threshold = missing_threshold
@@ -573,11 +603,7 @@ class FeatureSelector:
             len(self.ops['missing']), self.missing_threshold))
 
     def identify_single_unique(self):
-        """
-        Finds features with only a single unique value. NaNs do not count as a unique value.
-        :return:
-         self : object
-            Nothing but attributes
+        """Finds features with only a single unique value. NaNs do not count as a unique value.
         """
 
         # Calculate the unique counts in each column
@@ -601,16 +627,14 @@ class FeatureSelector:
         print('%d features with a single unique value.\n' % len(self.ops['single_unique']))
 
     def identify_high_cardinality(self, max_card=1000):
-        """
-        Finds the categorical columns with more than max_card unique values
+        """Finds the categorical columns with more than max_card unique values
         (might be relevant for GLM or if the categorical column has as many
         levels than rows in the data set)
 
-        :param max_card: int
+        Parameters
+        ----------
+        max_card: int
             the maximum number of the unique values
-        :return:
-         self : object
-            Nothing but attributes
         """
 
         col_names = list(self.data)
@@ -638,22 +662,23 @@ class FeatureSelector:
         )
 
     def encode_cat_var(self, df=None, col_excl=None, return_cat=False):
-        """
-        Categorical encoding (as integer). Automatically detect the non-numerical columns,
+        """Categorical encoding (as integer). Automatically detect the non-numerical columns,
         save the index and name of those columns, encode them as integer,
         save the direct and inverse mappers as dictionaries.
         Return the data-set with the encoded columns with a data type either int or pandas categorical.
 
-
-        :param df: pd.DataFrame
+        Parameters
+        ----------
+        df: pd.DataFrame
             the dataset
-        :param col_excl: list of str, default=None
+        col_excl: list of str, default=None
             the list of columns names not being encoded (e.g. the ID column)
-        :param return_cat: Boolean, default=False
+        return_cat: Boolean, default=False
             wether or not return pandas categorical (if false --> integer)
 
-        :return:
-         df: pd.DataFrame
+        Returns
+        -------
+        df: pd.DataFrame
             the dataframe with encoded columns
         """
         if df is None:
@@ -662,20 +687,19 @@ class FeatureSelector:
         else:
             is_external_data = True
 
-        df, cat_var_df, inv_mapper, mapper = cat_var(df, col_excl=col_excl, return_cat=return_cat)
+        df_enc, cat_var_df, inv_mapper, mapper = cat_var(df, col_excl=col_excl, return_cat=return_cat)
 
         self.cat_var_df = cat_var_df
         self.mapper = inv_mapper
 
         if not is_external_data:
             self.encoded = True
-            self.data = df
+            self.data = df_enc
 
-        return df
+        return df_enc
 
     def identify_collinear(self, correlation_threshold, encode=False, method='association'):
-        """
-        Finds collinear features based on the correlation coefficient between features.
+        """Finds collinear features based on the correlation coefficient between features.
         For each pair of features with a correlation coefficient greather than `correlation_threshold`,
         only one of the pair is identified for removal.
 
@@ -684,7 +708,6 @@ class FeatureSelector:
 
         Parameters
         ----------
-
         correlation_threshold : float between 0 and 1
             Value of the Pearson correlation cofficient for identifying correlation features
 
@@ -784,21 +807,18 @@ class FeatureSelector:
             len(self.ops['collinear']), self.correlation_threshold))
 
     def identify_zero_importance(self, task, eval_metric=None, objective=None,
-                                 n_iterations=10, early_stopping=True, zero_as_missing=False):
-        """
-
-        Identify the features with zero importance according to a gradient boosting machine.
+                                 n_iterations=10, use_early_stopping=True, zero_as_missing=False):
+        """Identify the features with zero importance according to a gradient boosting machine.
         The gbm can be trained with early stopping using a utils set to prevent overfitting.
         The feature importances are averaged over `n_iterations` to reduce variance.
 
         Uses the LightGBM implementation (http://lightgbm.readthedocs.io/en/latest/index.html)
 
         Parameters
-        --------
-
+        ----------
         eval_metric : string
             Evaluation metric to use for the gradient boosting machine for early stopping. Must be
-            provided if `early_stopping` is True
+            provided if `use_early_stopping` is True
 
         objective : string
             define the LGBM specific objective if any
@@ -809,7 +829,7 @@ class FeatureSelector:
         n_iterations : int, default = 10
             Number of iterations to train the gradient boosting machine
 
-        early_stopping : boolean, default = True
+        use_early_stopping : boolean, default = True
             Whether or not to use early stopping with a utils set when training
 
         zero_as_missing : Bool, default = False
@@ -817,8 +837,7 @@ class FeatureSelector:
 
 
         Notes
-        --------
-
+        -----
         - Features are one-hot encoded to handle the categorical variables before training.
         - The gbm is not optimized for any particular task and might need some hyperparameter tuning
         - Feature importances, including zero importance features,
@@ -826,7 +845,7 @@ class FeatureSelector:
 
         """
 
-        if early_stopping and eval_metric is None:
+        if use_early_stopping and eval_metric is None:
             raise ValueError("""eval metric must be provided with early stopping. 
             Examples include "auc" for classification or "l2" for regression.""")
 
@@ -844,6 +863,11 @@ class FeatureSelector:
         # Convert to np array
         features = np.array(features)
         labels = np.array(self.labels).reshape((-1,))
+        
+        if self.weight is not None:
+            weights = np.array(self.weight).reshape((-1,))
+        else:
+            weights = np.ones_like(labels)
 
         # Empty array for feature importances
         feature_importance_values = np.zeros(len(feature_names))
@@ -855,6 +879,7 @@ class FeatureSelector:
             progress_bar.set_description('Iteration nb: {0:<3}'.format(_))
 
             if task == 'classification':
+                self.stratified = True
                 model = lgb.LGBMClassifier(n_estimators=1000, learning_rate=0.05, verbose=-1,
                                            zero_as_missing=zero_as_missing)
 
@@ -869,22 +894,35 @@ class FeatureSelector:
                 model.set_params(**{'objective': objective})
 
             # If training using early stopping need a utils set
-            if early_stopping:
-                train_idx = random.sample(range(features.shape[0]), k=int(features.shape[0] * .85))
-                mask = np.zeros(features.shape[0], dtype=bool)
-                mask[train_idx] = True
-                weight = self.weight
-                train_features = features[mask]
-                valid_features = features[~mask]  # features.drop(features.index[train_idx])
-                train_labels = labels[mask]
-                valid_labels = labels[~mask]
-                train_weight = weight[mask]
-                valid_weight = weight[~mask]
+            if use_early_stopping:
+                if self.stratified:
+                    rs = StratifiedShuffleSplit(n_splits=1, test_size=.20, random_state=42)
+                    for train_index, test_index in rs.split(features, labels):
+                        valid_features, valid_labels, valid_weight = features[test_index, :], labels[test_index], weights[test_index] 
+                        train_features, train_labels, train_weight = features[train_index, :], labels[train_index], weights[train_index]
+                else:
+                    rs = ShuffleSplit(n_splits=1, test_size=.20, random_state=42)
+                    for train_index, test_index in rs.split(features):
+                        valid_features, valid_labels, valid_weight = features[test_index, :], labels[test_index], weights[test_index] 
+                        train_features, train_labels, train_weight = features[train_index, :], labels[train_index], weights[train_index]
+                
+                # train_idx = random.sample(range(features.shape[0]), k=int(features.shape[0] * .85))
+                # mask = np.zeros(features.shape[0], dtype=bool)
+                # mask[train_idx] = True
+                # weight = self.weight
+                # train_features = features[mask]
+                # valid_features = features[~mask]  # features.drop(features.index[train_idx])
+                # train_labels = labels[mask]
+                # valid_labels = labels[~mask]
+                # train_weight = weight[mask]
+                # valid_weight = weight[~mask]
 
                 # Train the model with early stopping
-                model.fit(train_features, train_labels, eval_metric=eval_metric,
+                model.fit(train_features, train_labels, 
+                          eval_metric=eval_metric,
                           eval_set=[(valid_features, valid_labels)],
-                          early_stopping_rounds=100, verbose=-1, sample_weight=train_weight,
+                          sample_weight=train_weight,
+                          callbacks=[early_stopping(20, False, False)],
                           eval_sample_weight=[valid_weight])
                 # pimp cool but too slow
                 # perm_imp =  permutation_importance(
@@ -893,7 +931,17 @@ class FeatureSelector:
                 # perm_imp = perm_imp.importances_mean
 
                 shap_matrix = model.predict(valid_features, pred_contrib=True)
-                shap_imp = np.mean(np.abs(shap_matrix[:, :-1]), axis=0)
+                # the dim changed in lightGBM 3
+                # X_SHAP_values array-like of shape = [n_samples, n_features + 1] or 
+                # shape = [n_samples, (n_features + 1) * n_classes] or list with n_classes length of such objects
+                if task == 'regression':
+                    shap_imp = np.mean(np.abs(shap_matrix[:, :-1]), axis=0)
+                elif task == 'classification':
+                    n_features = valid_features.shape[1]
+                    shap_matrix = np.delete(shap_matrix, list(range(n_features, shap_matrix.shape[1], n_features+1)), axis=1)
+                    shap_imp = np.mean(np.abs(shap_matrix), axis=0)
+                else:
+                    raise ValueError('Task must be either "classification" or "regression"')
 
                 # Clean up memory
                 del train_features, train_labels, valid_features, valid_labels
@@ -938,14 +986,13 @@ class FeatureSelector:
         print('\n%d features with zero importance after encoding.\n' % len(self.ops['zero_importance']))
 
     def identify_low_importance(self, cumulative_importance):
-        """
-        Finds the lowest importance features not needed to account for `cumulative_importance` fraction
+        """Finds the lowest importance features not needed to account for `cumulative_importance` fraction
         of the total feature importance from the gradient boosting machine. As an example, if cumulative
         importance is set to 0.95, this will retain only the most important features needed to
         reach 95% of the total feature importance. The identified features are those not needed.
 
         Parameters
-        --------
+        ----------
         cumulative_importance : float between 0 and 1
             The fraction of cumulative importance to account for
 
@@ -981,12 +1028,10 @@ class FeatureSelector:
               'importance of %0.2f.\n' % (len(self.ops['low_importance']), self.cumulative_importance))
 
     def identify_all(self, selection_params):
-        """
-        Use all five of the methods to identify features to remove.
+        """Use all five of the methods to identify features to remove.
 
         Parameters
-        --------
-
+        ----------
         selection_params : dict Parameters to use in the five feature selection methhods.
         Params must contain the keys
         ['patterns', 'missing_threshold', 'max_card','correlation_threshold', 'eval_metric', 'task',
@@ -1033,27 +1078,25 @@ class FeatureSelector:
 
         return list(self.all_identified)
 
-    def remove(self, methods):
-        """
-        Remove the features from the data according to the specified methods.
+    def remove(self, methods='all'):
+        """Remove the features from the data according to the specified methods.
 
         Parameters
-        --------
-            methods : 'all' or list of methods
-                If methods == 'all', any methods that have identified features will be used
-                Otherwise, only the specified methods will be used.
-                Can be one of ['missing', 'single_unique', 'collinear', 'zero_importance', 'low_importance']
-        Return
-        --------
-            data : dataframe
-                Dataframe with identified features removed
-
+        ----------
+        methods : str or list of str
+            If methods == 'all', any methods that have identified features will be used
+            Otherwise, only the specified methods will be used.
+            Can be one of ['missing', 'single_unique', 'collinear', 'zero_importance', 'low_importance']
+        Returns
+        -------
+        data : dataframe
+            Dataframe with identified features removed
 
         Notes
-        --------
-            - If feature importances are used, the one-hot encoded columns will be
+        -----
+        - If feature importances are used, the one-hot encoded columns will be
               added to the data (and then may be removed)
-            - Check the features that will be removed before transforming data!
+        - Check the features that will be removed before transforming data!
 
         """
 
@@ -1154,18 +1197,34 @@ class FeatureSelector:
         # plt.yscale("log", nonposy='clip')
 
     def plot_collinear(self, plot_all=False, size=1000):
-        """
-        Heatmap of the correlation values. If plot_all = True plots all the correlations otherwise
-        plots only those features that have a correlation above the threshold
+        """Heatmap of the correlation values, reduced or not
 
+        Parameters
+        ----------
+        plot_all : bool, optional
+            If plot_all = True plots all the correlations otherwise
+            plots only those features that have a correlation above the threshold
+        size : int, optional
+            figure size in in pixels, by default 1000
+
+        Returns
+        -------
+        panel_layout : panel
+            the interactive plot
+            
         Notes
-        --------
-            - Not all of the plotted correlations are above the threshold because this plots
-            all the variables that have been idenfitied as having even one correlation above the threshold
-            - The features on the x-axis are those that will be removed. The features on the y-axis
-            are the correlated features with those on the x-axis
+        -----
+        - Not all of the plotted correlations are above the threshold because this plots
+        all the variables that have been idenfitied as having even one correlation above the threshold
+        - The features on the x-axis are those that will be removed. The features on the y-axis
+        are the correlated features with those on the x-axis
 
         Code adapted from https://seaborn.pydata.org/examples/many_pairwise_correlations.html
+
+        Raises
+        ------
+        NotImplementedError
+            if correlation matrix has not been computed yet
         """
 
         if self.record_collinear is None:
@@ -1219,20 +1278,26 @@ class FeatureSelector:
 
 
     def plot_feature_importances(self, plot_n=15, threshold=None, figsize=None):
-        """
-        Plots `plot_n` most important features and the cumulative importance of features.
+        """Plots `plot_n` most important features and the cumulative importance of features.
         If `threshold` is provided, prints the number of features needed to reach `threshold`
         cumulative importance.
 
         Parameters
-        --------
-
+        ----------
         plot_n : int, default = 15
             Number of most important features to plot. Defaults to 15 or the maximum
             number of features whichever is smaller
 
         threshold : float, between 0 and 1 default = None
             Threshold for printing information about cumulative importances
+            
+        figsize : tuple of int
+            The rendered size as a percentage size
+            
+        Returns
+        -------
+        hv.plot
+            the feature importances holoviews object 
 
         """
 
